"""
cophi.utils
~~~~~~~~~~~

This module implements low-level helper functions.
"""

import collections
import itertools

import pandas as pd
import regex as re


def construct_ngrams(tokens, n=2, sep=" "):
    """
    Parameters:
        tokens (list): The tokenized document.
        n (int): Number of tokens per ngram.
        sep (str): Separator between tokens within an ngram.
    """
    return (sep.join(ngram)
            for ngram in zip(*(itertools.islice(i, token, None)
                             for token, i in enumerate(itertools.tee(tokens,
                                                                     n)))))


def find_tokens(document, token_pattern=r"\p{L}+\p{Connector_Punctuation}?\p{L}+", maximum=None):
    """
    Parameters:
        document (str): The text of a document.
        token_pattern (str): Regex pattern for one token.
        maximum (int): Stop tokenizing after that much tokens.
    """
    count = 0
    for match in re.compile(token_pattern).finditer(document):
        count += 1
        yield match.group(0)
        if maximum is not None and count >= maximum:
            return


def lowercase_tokens(tokens):
    """
    Parameters:
        tokens (list): The tokenized document.
    """
    return [token.lower() for token in tokens]


def segment_fuzzy(paragraphs, size=1000, tolerance=0.05):
    """Segment a string, respecting paragraphs.

    Parameters:
        paragraphs: Paragraphs of a document as separated entities.
        size: The target length of each segment in tokens.
        tolerance: How much may the actual segment size differ from the `size`?
            If ``0 < tolerance < 1``, this is interpreted as a fraction
            of the `size`, otherwise it is interpreted as an absolute number.
            If ``tolerance < 0``, paragraphs are never split apart.
    """
    if tolerance > 0 and tolerance < 1:
        tolerance = round(size * tolerance)
    current_segment = []
    carry = None
    current_size = 0
    doc_iter = iter(paragraphs)
    try:
        while True:
            chunk = list(carry if carry else next(doc_iter))
            carry = None
            current_segment.append(chunk)
            current_size += len(chunk)
            if current_size >= size:
                too_long = current_size - size
                too_short = size - (current_size - len(chunk))
                if tolerance >= 0 and min(too_long, too_short) > tolerance:
                    chunk_part0 = chunk[:-too_long]
                    carry = chunk[-too_long:]
                    current_segment[-1] = chunk_part0
                elif too_long >= too_short:
                    carry = current_segment.pop()
                yield current_segment
                current_segment = []
                current_size = 0
    except StopIteration:
        pass
    if current_segment:
        yield current_segment


def _parameter(tokens, measure):
    """Count types, tokens and occuring frequencies.

    Parameters:
        tokens (list): The tokenized document.
        measure (str): Complexity measure you need the results to.
    """
    if measure in {"sichel_s", "michea_m"}:
        bow = collections.Counter(tokens)
        freq_spectrum = collections.Counter(bow.values())
        return {"num_types": len(bow), "freq_spectrum": freq_spectrum}
    elif measure in {"honore_h", "herdan_vm", "orlov_z"}:
        bow = collections.Counter(tokens)
        freq_spectrum = collections.Counter(bow.values())
        return {"num_types": len(bow), "num_tokens": len(tokens),
                "freq_spectrum": pd.Series(freq_spectrum)}
    elif measure in {"entropy", "yule_k", "simpson_d"}:
        bow = collections.Counter(tokens)
        freq_spectrum = collections.Counter(bow.values())
        return {"num_tokens": len(tokens),
                "freq_spectrum": pd.Series(freq_spectrum)}
    else:
        return {"num_types": len(set(tokens)), "num_tokens": len(tokens)}



"""
cophi.api
~~~~~~~~~

This module implements the high-level API.
"""

import logging
import pathlib
import uuid
import pandas as pd
import cophi.model


logger = logging.getLogger(__name__)


def document(filepath, **kwargs):
    """Read a text file and create a Document object.

    Parameter:
        filepath (str): Path to the text file.
        title (str): Describing title for the document. (optional).
        lowercase (bool): If True, writes all letters in lowercase (optional).
        n (int): Number of tokens per ngram (optional).
        token_pattern (str): Regex pattern for one token (optional).
        maximum (int): Stop tokenizing after that much tokens (optional).

    Returns:
        A Document object.
    """
    textfile = cophi.model.Textfile(filepath)
    return cophi.model.Document(textfile.content, **kwargs)


def corpus(directory, filepath_pattern="*", treat_as=None, encoding="utf-8",
           lowercase=True, n=None, token_pattern=r"\p{L}+\p{P}?\p{L}+",
           maximum=None, metadata=False):
    """Pipe a collection of text files and create a Corpus object.

    Parameters:
        directory (str): Path to the corpus directory.
        filepath_pattern (str): Glob pattern for text files (optional).
        treat_as (str): Treat text files like .txt or .xml (optional).
        encoding (str): Encoding to use for UTF when reading (optional).
        lowercase (bool): If True, writes all letters in lowercase (optional).
        n (int): Number of tokens per ngram (optional).
        token_pattern (str): Regex pattern for one token (optional).
        maximum (int): Stop tokenizing after that much tokens (optional).
        metadata (bool): Extract metadata from filenames (optional).

    Returns:
        A Corpus model object and optionally a Metadata object.
    """
    if not isinstance(directory, pathlib.Path):
        directory = pathlib.Path(directory)
    filepaths = directory.rglob(filepath_pattern)

    def lazy_reading(filepaths):
        for filepath in filepaths:
            if filepath.is_file() and ".git" not in str(filepath):
                yield cophi.model.Textfile(filepath, treat_as, encoding)

    if metadata:
        metadata_ = cophi.model.Metadata()
    documents = pd.Series()
    for textfile in lazy_reading(filepaths):
        logger.info("Processing '{}' ...".format(textfile.title))
        title = str(uuid.uuid1()) if metadata else textfile.title
        text = textfile.content
        document = cophi.model.Document(text,
                                        title,
                                        token_pattern,
                                        lowercase,
                                        n,
                                        maximum)
        documents[title] = document
        if metadata:
            metadata_ = metadata_.append({"uuid": title,
                                          "filepath": textfile.filepath,
                                          "parent": textfile.parent,
                                          "title": textfile.title,
                                          "suffix": textfile.filepath.suffix},
                                          ignore_index=True)
    logger.info("Constructing Corpus object ...")
    if metadata:
        return cophi.model.Corpus(documents), metadata_
    else:
        return cophi.model.Corpus(documents)

def export(dtm, filepath, format="text"):
    """Export a document-term matrix.

    Parameters:
        dtm: A document-term matrix.
        filepath: Path to output file. Possibel values are `plaintext`/`text` or
            `svmlight`.
        format: File format.
    """
    if format.lower() in {"plaintext", "text"}:
        cophi.model.Corpus.plaintext(dtm, filepath)
    elif format.lower() in {"svmlight"}:
        cophi.model.Corpus.svmlight(dtm, filepath)
    else:
        raise ValueError("'{}' is no supported file format.".format(format))



"""
cophi.complexity
~~~~~~~~~~~~~~~~

This module implements measures that assess the linguistic
and stylistic complexity of (literary) texts.

:math:`N` is the Absolute number of tokens, and :math:`V` the
Absolute number of types. :math:`H` is the Absolute number of
types occuring only once (hapax legomena), :math:`D` is the
absolute number of types occuring twice (dislegomena).

This module was taken from
`here <https://github.com/tsproisl/Linguistic_and_Stylistic_Complexity>`_.
"""

import math
import numpy as np


# use num_types + num_tokens (int):

def ttr(num_types, num_tokens):
    r"""Calculate Type-Token Ratio (TTR).

    Used formula:
        .. math::
            TTR = \frac{V}{N}

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
    """
    return num_types / num_tokens


def guiraud_r(num_types, num_tokens):
    r"""Calculate Guiraud’s R (1954).

    Used formula:
        .. math::
            R = \frac{V}{\sqrt{N}}

    Parameters:
            num_types (int): Absolute number of types.
            num_tokens (int): Absolute number of tokens.
    """
    return num_types / math.sqrt(num_tokens)


def herdan_c(num_types, num_tokens):
    r"""Calculate Herdan’s C (1960, 1964).

    Used formula:
        .. math::
            C = \frac{\log{V}}{\log{N}}

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
    """
    return math.log(num_types) / math.log(num_tokens)


def dugast_k(num_types, num_tokens):
    r"""Calculate Dugast’s k (1979).

    Used formula:
        .. math::
            k = \frac{\log{V}}{\log{\log{N}}}

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
    """
    return math.log(num_types) / math.log(math.log(num_tokens))


def maas_a2(num_types, num_tokens):
    r"""Calculate Maas’ a^2 (1972).

    Used formula:
        .. math::
            a^2 = \frac{\log{N} \; - \; \log{V}}{\log{N}^2}

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
    """
    return (math.log(num_tokens)
            - math.log(num_types)) / (math.log(num_tokens) ** 2)


def dugast_u(num_types, num_tokens):
    r"""Calculate Dugast’s U (1978, 1979).

    Used formula:
        .. math::
            U = \frac{\log{N^2}}{\log{N} \; - \; \log{V}}

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
    """
    return (math.log(num_tokens) ** 2) / (math.log(num_tokens)
                                          - math.log(num_types))


def tuldava_ln(num_types, num_tokens):
    r"""Calculate Tuldava’s LN (1977).

    Used formula:
        .. math::
            LN = \frac{1 \; - \; V^2}{V^2 \; \cdot \; \log{N}}

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
    """
    return (1 - (num_types ** 2)) / ((num_types ** 2) * math.log(num_tokens))


def brunet_w(num_types, num_tokens):
    r"""Calculate Brunet’s W (1978).

    Used formula:
        .. math::
            W = V^{V^{0.172}}

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
    """
    a = -0.172
    return num_tokens ** (num_types ** -a)


def cttr(num_types, num_tokens):
    r"""Calculate Carroll’s Corrected Type-Token Ration (CTTR) (1964).

    Used formula:
        .. math::
            CTTR = \frac{V}{\sqrt{2 \; \cdot \; N}}

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
    """
    return num_types / math.sqrt(2 * num_tokens)


def summer_s(num_types, num_tokens):
    r"""Calculate Summer’s S.

    Used formula:
        .. math::
            S = \frac{\log{\log{V}}}{\log{\log{N}}}

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
    """
    return math.log(math.log(num_types)) / math.log(math.log(num_tokens))


# use num_types + part of freq_spectrum:

def sichel_s(num_types, freq_spectrum):
    r"""Calculate Sichel’s S (1975).

    Used formula:
        .. math::
            S = \frac{D}{V}

    Parameters:
        num_types (int): Absolute number of types.
        freq_spectrum (dict): Counted occurring frequencies.
    """
    return freq_spectrum[2] / num_types


def michea_m(num_types, freq_spectrum):
    r"""Calculate Michéa’s M (1969, 1971).

    Used formula:
        .. math::
            M = \frac{V}{D}

    Parameters:
        num_types (int): Absolute number of types.
        freq_spectrum (dict): Counted occurring frequencies.
    """
    return num_types / freq_spectrum[2]


def honore_h(num_types, num_tokens, freq_spectrum):
    r"""Calculate Honoré’s H (1979).

    Used formula:
        .. math::
            H = 100 \cdot \frac{\log{N}}{1 - \frac{H}{V}}

    Parameters:
        num_types (int): Absolute number of types.
        freq_spectrum (dict): Counted occurring frequencies.
    """
    return 100 * (math.log(num_tokens)
                  / (1 - ((freq_spectrum[1]) / (num_types))))


# use num_tokens + freq_spectrum:

def entropy(num_tokens, freq_spectrum):
    r"""Calculate entropy S.

    Parameters:
        num_tokens (int): Absolute number of tokens.
        freq_spectrum (pd.Series): Counted occurring frequencies.
    """
    a = freq_spectrum.index.values / num_tokens
    b = - np.log(a)
    result = freq_spectrum.values * a * b
    return result.sum()


def yule_k(num_tokens, freq_spectrum):
    r"""Calculate Yule’s K (1944).

    Used formula:
        .. math::
            K = 10^4 \times \frac{(\sum_{X=1}^{X}{{f_X}X^2}) - N}{N^2}

    Parameters:
        num_tokens (int): Absolute number of tokens.
        freq_spectrum (pd.Series): Counted occurring frequencies.
    """
    a = freq_spectrum.index.values / num_tokens
    b = 1 / num_tokens
    return 10 ** 4 * ((freq_spectrum.values * a ** 2) - b).sum()


def simpson_d(num_tokens, freq_spectrum):
    """Calculate Simpson’s D.

    Parameters:
        num_tokens (int): Absolute number of tokens.
        freq_spectrum (pd.Series): Counted occurring frequencies.
    """
    a = freq_spectrum.values / num_tokens
    b = freq_spectrum.index.values - 1
    return (freq_spectrum.values * a * (b / (num_tokens - 1))).sum()


def herdan_vm(num_types, num_tokens, freq_spectrum):
    """Calculate Herdan’s VM (1955).

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
        freq_spectrum (pd.Series): Counted occurring frequencies.
    """
    a = freq_spectrum.index.values / num_tokens
    b = 1 / num_types
    return math.sqrt((freq_spectrum * a ** 2).sum() - b)


# use probabilistic models:

def orlov_z(num_tokens, num_types, freq_spectrum,
            max_iterations=100, min_tolerance=1):
    """Calculate Orlov’s Z (1983), approximated via Newton’s method.

    Parameters:
        num_types (int): Absolute number of types.
        num_tokens (int): Absolute number of tokens.
        freq_spectrum (dict): Counted occurring frequencies.
    """
    most_frequent = max(freq_spectrum)
    p_star = most_frequent / num_tokens
    z = num_tokens / 2
    for i in range(max_iterations):
        next_z = z - (_get_z(num_tokens,
                             num_types,
                             p_star, z) / _derivative(num_tokens,
                                                      num_types,
                                                      p_star,
                                                      z))
        abs_diff = abs(z - next_z)
        z = next_z
        if abs_diff <= min_tolerance:
            break
    return z


def _get_z(num_tokens, num_types, p_star, z):
    """Private function for :func:`orlov_z`.
    """
    return (((z / math.log(p_star * z))
            * (num_tokens / (num_tokens - z))
            * math.log(num_tokens / z) - num_types))


def _derivative(num_tokens, num_types, p_star, z):
    """Private function for :func:`orlov_z`.
    """
    return ((num_tokens
             * ((z - num_tokens)
             * math.log(p_star * z)
             + math.log(num_tokens / z)
             * (num_tokens * math.log(p_star * z) - num_tokens + z)))
            / (((num_tokens - z) ** 2) * (math.log(p_star * z) ** 2)))


# other:

def ci(results):
    """Calculate  the confidence interval for standardized TTR.

    Parameters:
        results (list): Bootstrapped TTRs.
    """
    return 1.96 * np.std(results) / math.sqrt(len(results))


def wrapper(measure):
    """A wrapper for all complexity functions.
    """
    measures = {"ttr": ttr,
                "guiraud_r": guiraud_r,
                "herdan_c": herdan_c,
                "dugast_k": dugast_k,
                "maas_a2": maas_a2,
                "dugast_u": dugast_u,
                "tuldava_ln": tuldava_ln,
                "brunet_w": brunet_w,
                "cttr": cttr,
                "summer_s": summer_s,
                "sichel_s": sichel_s,
                "michea_m": michea_m,
                "honore_h": honore_h,
                "entropy": entropy,
                "yule_k": yule_k,
                "simpson_d": simpson_d,
                "herdan_vm": herdan_vm,
                "orlov_z": orlov_z}
    return measures[measure]
